/*
 * ctapi.c linux version
 * 
 * ctapi for serial connection (including USB / BT, etc.)
 * extended for SICCT terminal handling 11.5.2011 PeWi
 * (c) G&W Instruments 2011 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Created on: 10.02.2011
 *      Author: steinbeck
 */

#include <stdio.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/sockios.h>
#include <linux/if.h>

#include "ctapi.h"
#include "sicct.h"
#include "sdd.h" 

#define INVALID_HANDLE_VALUE	-1
#define INVALID_SOCKET		-1

#define IB_SEQ		0x40	/* sequence number in case of I-Block */
#define IB_CHAIN	0x20
#define	RB_SEQ		0x10	/* sequence number in case of R-Block */
#define RB		0x80	/* R-Block ID */
#define WTX_RSP		0xE3

#define MAX_TX_LEN	254	/* 254 default, make shorter to test chaining */
#define CARD_TYPE_KVK	0
#define CARD_TYPE_EGK	1
#define CARD_NONE	-1

#define CONFIG_NONE	0
#define CONFIG_IP	1
#define CONFIG_MAC	2
#define CONFIG_COM	3

#undef  DEBUG
// #define DEBUG

#ifdef DEBUG
# define DBG(fmt, args...)       printf(fmt, ##args)
#else
# define DBG(fmt, args...)
#endif

static const unsigned char ask_resync[] = {0x12, 0xC0, 0x00, 0xD2};
static const unsigned char read_data_cmd[] = {0x00,0xb0, 0x00, 0x00, 0x00} ;
static const unsigned char reset_ct[] = {0x20, 0x11, 0x00, 0x00} ;

pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
struct terminal_st {
	unsigned short 		ctn ;
	unsigned short  	pn ;
	int			last_wtx ;
	int			port_fd ;
	int			Socket ;
	struct sockaddr_in	sa ;
	unsigned short 		tx_seq ;
	unsigned char		iseq ;
	char			devname[20];
	char			ip_addr[20];
	char			mac_addr[20];
	int			config_type;
	struct terminal_st	*next ;
	struct termios old_tm;
	struct termios tm ;
};

static struct terminal_st *termlist = NULL ;

void dump(int n, unsigned char * data)
{
#ifdef DEBUG
	int i, len ;

	while(n) {
		if(n >= 16) len = 16 ;
		else len = n ;

		for(i=0 ; i< len ; i++) printf("%.2x ", data[i]) ;
		for( ; i<16 ; i++)  printf("   ") ;
		printf(" ") ;
		for(i=0 ; i< len ; i++) {
			if( (data[i] < 0x20) || (data[i] > 127) ) printf(".") ;
			else printf("%c", data[i]) ;
		}
		printf("\n") ;
		n -= len ;
		data += len ;
	}
#endif
}

int print_bc(int n, unsigned char *data)
{
	int ret = 0 ;

	if( data[0] != 0x20 ) return ret  ;

	switch(data[1]) {

		case 0x11:
			if(data[2] == 0) {
				DBG("Reset CT\n") ;
			} else {
				DBG("Reset ICC %d Response Format %d\n", (int) data[2], (int) data[3] ) ;
			}
			ret = 1 ;
			break ;

		case 0x12:
			DBG("Request ICC %d Qualifier=0x%.2x\n", (int) data[2], (int) data[3]) ;
			if(n > 4) {
				dump(n, data) ;
			}
			ret = 1 ;
			break ;

		case 0x13:
			DBG("Get Status Unit %d, Qualifier=0x%.2x\n", (int)data[2], (int)data[3]) ;
			ret = 1 ;
			break ;

		case 0x15:
			DBG("Eject ICC %d, Qualifier=0x%.2x\n", (int)data[2], (int)data[3]) ;
			ret = 1 ;
			break ;

		default:
			break ;
	}

	return ret ;
}

static unsigned char calc_csum(int len, unsigned char *data)
{
	unsigned char csum = 0 ;

	while(len--) csum = csum ^ *data++ ;

	return csum ;
}

int add_terminal(struct terminal_st *ele)
{
	struct terminal_st *q  ;

	pthread_mutex_lock(&mtx) ;
	q = termlist ;
	if(q) {
		while(q->next) q=q->next ;
		q->next = ele ;
	} else {
		termlist = ele ;
	}
	pthread_mutex_unlock(&mtx) ;
	return 0 ;
}

int remove_terminal(unsigned short ctn)
{
	struct terminal_st *q, **pq ;
	int ret = -1 ; /* assume not found */

	pthread_mutex_lock(&mtx) ;
	q = termlist ;
	pq = &termlist ;
	while(q) {
		if(q->ctn == ctn) {
			*pq = q->next ;
			free(q) ;
			ret = 0 ;
			break ;
		}
		pq = &q->next ;
		q  = q->next ;
	}
	pthread_mutex_unlock(&mtx) ;
	return ret ;
}

struct terminal_st *get_terminal(unsigned short ctn)
{
	struct terminal_st *q , *term ;

	term = NULL ;

	pthread_mutex_lock(&mtx) ;
	q = termlist ;
	while(q) {
		if(q->ctn == ctn) {
			term = q  ;
			break ;
		}
		q = q->next ;
	}
	pthread_mutex_unlock(&mtx) ;

	return term ;
}

struct terminal_st *new_terminal(void)
{
	struct terminal_st *term ;

	term = (struct terminal_st *) malloc(sizeof(struct terminal_st)) ;
	if(!term) return NULL ;

	memset(term, 0, sizeof(struct terminal_st)) ;
	term->port_fd	= INVALID_HANDLE_VALUE ;
	term->Socket	= INVALID_SOCKET ;
	return term ;
} ;

//---------------------------------------------------------------------------

/* network buffer acccess */
unsigned short get_nbufs(unsigned char *buf)
{
	unsigned short val ;

	val = (unsigned short) *(buf+1) ;
	val <<= 8 ;
	val |= (unsigned short) *(buf) ;
	return ntohs(val) ;
}

unsigned long get_nbufl(unsigned char *buf)
{
	unsigned long val ;

	val = (unsigned long) *(buf+3) ;
	val <<= 8 ;
	val |= (unsigned long) *(buf+2) ;
	val <<= 8 ;
	val |= (unsigned long) *(buf+1) ;
	val <<= 8 ;
	val |= (unsigned long) *(buf) ;
	return ntohl(val) ;
}

void set_nbufs(unsigned char *buf, unsigned short hval)
{
	unsigned short nval = htons(hval) ;

	*buf     = (unsigned char) (nval & 0xff) ;
	nval >>= 8 ;
	*(buf+1) = (unsigned char) (nval & 0xff) ;
}

void set_nbufl(unsigned char *buf, unsigned long hval)
{
	unsigned long nval = htonl(hval) ;

	*buf     = (unsigned char) (nval & 0xff) ;
	nval >>= 8 ;
	*(buf+1) = (unsigned char) (nval & 0xff) ;
	nval >>= 8 ;
	*(buf+2) = (unsigned char) (nval & 0xff) ;
	nval >>= 8 ;
	*(buf+3) = (unsigned char) (nval & 0xff) ;
}

//---------------------------------------------------------------------------
// SICCT Terminal Search 
//---------------------------------------------------------------------------
#define SDC_VERSION 0x0114

static int build_sdd_request(unsigned char *buf, int len, struct in_addr *ip)
{
	int pos = 0 ;

	buf[pos++] = SDD_REQUEST ;
	buf[pos++] = 0 ; 	/* total len */
	buf[pos++] = TAG_SDD_VERSION ;
	buf[pos++] = 2 ;
	set_nbufs(&buf[pos], SDC_VERSION) ;
	pos += 2 ;

	buf[pos++] = TAG_SENDER_IP ;
	buf[pos++] = (unsigned char) (sizeof(struct in_addr) & 0xff) ;
	memcpy(&buf[pos], ip, sizeof(struct in_addr)) ;
	pos += sizeof(struct in_addr) ;

	buf[pos++] = TAG_SICCT_PORT ;
	buf[pos++] = 2 ;
	set_nbufs(&buf[pos], SDD_PORT) ;
	pos += 2 ;

	buf[1] = (unsigned char)((pos-2) & 0x7f) ;
	
	return pos ;
}

static int check_sdd_response(unsigned char *buf, int len, unsigned char *mac)
{
	int pos = 0 ;
	int dlen = 0 ;
	unsigned char *pc, *pend ;
	char tbuf[64] ;
	int ret = 0; 

	if(buf[pos++] != SDD_RESPONSE) return 0 ;
	dlen = (int)buf[pos++] ;
	if(dlen != (len-2)) return 0 ; 

	pc = &buf[pos] ;
	pend = pc+dlen ; 
	while(pc < pend) {
		switch(*pc++) {
			case TAG_SDD_VERSION:
				if(*pc++ != 2) return -1 ; 
				DBG("Response Version = %2d.%2d\n", (int)pc[0], (int)pc[1]) ; 
				pc+=2 ;
				break ;
			case TAG_SENDER_IP: 
				if(*pc++ != 4) return -1 ; 
				DBG("Response IP: = %3d.%3d.%3d.%3d\n",
				     (int)pc[0], (int)pc[1], (int)pc[2], (int)pc[3]) ; 
				pc+=4 ;
				break ;
			case TAG_SERVER_MAC:
				if(*pc++ != 6) return -1 ; 
				DBG("Response MAC: = %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
				     (int)pc[0], (int)pc[1], (int)pc[2], (int)pc[3], (int)pc[4], (int)pc[5]) ;
				DBG("Search MAC: = %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
				     (int)mac[0], (int)mac[1], (int)mac[2], (int)mac[3], (int)mac[4], (int)mac[5]) ;
				if( 0 == memcmp(pc, mac, 6) ) ret = 1 ; 
				pc+=6 ;
				break ; 
			case TAG_SICCT_NAME:
				dlen = (int)*pc++ ;
				memset(tbuf, 0, 64)  ;
				memcpy(tbuf, pc, dlen<63 ? dlen:63 ) ;
				DBG("Response Name (l=%d) %s\n", dlen, tbuf) ;  
				pc += dlen ;
				break ; 
			case TAG_SICCT_PORT: 
				if(*pc++ != 2) return -1 ;
				DBG("Response Port %d\n", (int) get_nbufs(pc) ) ; 
				pc+=2 ;
				break ;
			case TAG_SICCT_SECURITY:
				if(*pc++ != 3) return -1 ;
				DBG("Response Security %2x %2x %2x\n", (int)pc[0], (int)pc[1], (int)pc[2]) ;
				pc+=3 ;
				break ;
			default:
				DBG("unknown request tag %d\n", (int)(--*pc) ) ;
				pc++ ; 
				dlen = (int)*pc++ ;
				pc += dlen ; 
				break ; 
		}
	}
	return ret ; 
}

int get_my_ips(int sk, struct in_addr *ips, int nmax)
{
	struct ifreq ifreqs[8];
	struct ifconf ifconf;
	int  nifaces, ret ;
	int cnt = 0; 

	/* get interfaces and states for socket */
 	memset(&ifconf,0,sizeof(ifconf));
	ifconf.ifc_buf = (char*) (ifreqs);
	ifconf.ifc_len = sizeof(ifreqs);
	if((ret = ioctl(sk, SIOCGIFCONF , (char*) &ifconf)) < 0 ) {
		DBG("Getting interfaces failed\n");
		nifaces = 0 ;
	} else {
		int i ;
		nifaces = ifconf.ifc_len/sizeof(struct ifreq);
		for(i=0 ; i<nifaces ; i++) {
			ret = ioctl(sk, SIOCGIFFLAGS, (char*) &ifreqs[i]) ;
			if(ret < 0) {
				DBG("failed getting flags for iface %s\n", ifreqs[i].ifr_name) ;
			} else {
				DBG("iface %d %s\t flags 0x%x\t %s\t %s\n",
				     i, ifreqs[i].ifr_name, ifreqs[i].ifr_flags,
				     ifreqs[i].ifr_flags & IFF_UP ? "UP":"",
				     ifreqs[i].ifr_flags & IFF_RUNNING ? "RUNNING":"") ;
				/* skip local */
				if(0 == strncmp("lo", ifreqs[i].ifr_name, 3)) continue;

				/* only use interfaces which are UP */
				if( ! (ifreqs[i].ifr_flags & IFF_RUNNING)  ) continue; 

				/* use only if we have an HW address */
				ret = ioctl(sk, SIOCGIFHWADDR, &ifreqs[i]) ;
				if(ret < 0) {
					DBG("Failed getting HWADDR for %s\n", ifreqs[i].ifr_name) ;
					continue; 
				}
				ret = ioctl(sk, SIOCGIFADDR, &ifreqs[i]) ;
				if(ret < 0)  {
					DBG("Failed getting IP for %s\n", ifreqs[i].ifr_name) ;
				} else {
					struct sockaddr_in *sa=(struct sockaddr_in *)&(ifreqs[i].ifr_addr);
					DBG("IP : %s\n",inet_ntoa(sa->sin_addr)) ;
					memcpy(&ips[cnt], &sa->sin_addr, sizeof(struct in_addr));
					cnt++;
					if(cnt >= nmax) break; 
				}
			}
		}
	}

	return cnt;
}


static int find_sicct_term(unsigned char *mac, char *ip)
{
	int num_ips ;
	struct in_addr ips[8] ;
	int i ;
	int sd ;
	int iopt ;
	struct sockaddr_in addr;
	unsigned char buf[256] ;
	int ntx, nrx ;
	int res ;
	int ret = 0 ;
	struct pollfd psd ;

	sd = socket(AF_INET, SOCK_DGRAM, 0);

	if(sd == INVALID_SOCKET) {
		DBG("Create Socket for SDD failed\n") ;
		return 0 ; 
	}
	iopt = 1;
	if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &iopt, sizeof(iopt)) < 0) {
		DBG("setsockopt broadcast failed\n");
		close(sd) ;
		return 0  ; 
	}
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &iopt, sizeof(iopt)) < 0) {
		DBG("setsockopt reuse addr failed\n");
		close(sd) ;
		return 0  ; 
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(SDD_PORT);
	addr.sin_addr.s_addr = INADDR_ANY;
	if ( bind(sd, (struct sockaddr*)&addr, sizeof(addr)) != 0 ) {
		DBG("failed to bind to socket\n") ;
		close(sd) ;
		return 0 ; 
	}

	num_ips = get_my_ips(sd, ips, 8);
	DBG("Got %d local IPs\n", num_ips) ;

	for(i=0 ; i<num_ips; i++) {
 		DBG("searching port %s\n", inet_ntoa(ips[i])) ;
		ntx = build_sdd_request(buf, 256, &ips[i]) ;
		addr.sin_port = htons(SDD_PORT) ; 
		addr.sin_addr.s_addr = htonl(INADDR_BROADCAST) ;
		sendto(sd, buf, ntx, 0, (struct sockaddr*)&addr, sizeof(addr));
		while(1) {
			psd.fd = sd ;
			psd.events = POLLIN | POLLPRI | POLLHUP | POLLERR ;
			res = poll(&psd, 1, 5000) ;
			if(res <= 0) {
				DBG("Poll timeout or error %d\n", res) ;
				break  ; 
			}
			if(psd.revents & (POLLIN | POLLPRI)) {
				socklen_t addr_len = sizeof(struct sockaddr_in) ; 
				nrx = recvfrom(sd, buf, 256, 0, (struct sockaddr*)&addr,  &addr_len);
				DBG("got response from %s len=%d\n", inet_ntoa(addr.sin_addr), nrx) ;
				dump(nrx, buf) ; 
				if( check_sdd_response(buf, nrx, mac) ) {
					sprintf(ip,"%s",inet_ntoa(addr.sin_addr))  ;
					ret = 1 ;
					break ; 
				}
			}
		}
		if(ret) break ; 
	}
	close(sd) ; 

	return ret ; 	
}

//---------------------------------------------------------------------------

int get_config(struct terminal_st *terminal )
{
	FILE *conf;
	char *read_buffer;
	char *parameter_value=NULL;
	char *interface_value=NULL;
	char *values= NULL;
	char parameter[32];
	read_buffer = (char*) malloc(256);
	int config_type=CONFIG_NONE;

	conf = fopen("/etc/ctapi.conf","rb");
	if (conf == NULL) {
		DBG("Could not open config file /etc/ctapi.conf \n" );
		return -1;
	}
	while( !feof(conf) ) {

		fgets(read_buffer, 255, conf);

		if ((values = strchr(read_buffer, '\r')) != 0)
			*values = 0;
		if ((values = strchr(read_buffer, '\n')) != 0)
			*values = 0;
		values=NULL;

		if ( (strlen(read_buffer)) == 0) {
			continue;
		}
		if (strstr(read_buffer,"#") != NULL) {
			continue;
		}

		interface_value = strchr(read_buffer, ':');
		if (interface_value != NULL) {
			strncpy(parameter, read_buffer, interface_value-read_buffer);
			parameter[interface_value-read_buffer] = '\0';
			interface_value+=1;
		}
		else {
			continue;
		}
		if (strcmp(parameter,"interface") == 0) {
			char *ptr_parameter = NULL;
			ptr_parameter = strtok(interface_value, ",");
			while(ptr_parameter != NULL) {

				parameter_value = strchr(ptr_parameter, '=');
				if (parameter_value != NULL) {
					/*we have a parameter*/
					strncpy(parameter, ptr_parameter, parameter_value-ptr_parameter);
					parameter[parameter_value-ptr_parameter] = '\0';
					parameter_value+=1;
				}
				else {
					continue;
				}

				if (strcmp(parameter,"devname") == 0 ) {
					strcpy(terminal->devname, parameter_value);
					config_type=CONFIG_COM;
				}
				if (strcmp(parameter,"ip_address") == 0 ) {
					strcpy(terminal->ip_addr, parameter_value);
					config_type=CONFIG_IP;
				}
				if (strcmp(parameter,"mac_address") == 0 ) {
					strcpy(terminal->mac_addr, parameter_value);
					config_type=CONFIG_MAC;
				}
				if (strcmp(parameter,"portnumber") == 0 ) {
					int port = atoi(parameter_value);
					if(port == terminal->pn && config_type != CONFIG_NONE) {
						DBG("found config: devname %s, ip_addr %s, mac_addr %s, "
						"portnum %d\n",terminal->devname,
						terminal->ip_addr ,
						terminal->mac_addr ,
						terminal->pn);
						ptr_parameter=NULL;
						memset(read_buffer,'\0',255);
						free(read_buffer);
						fclose(conf);
						parameter_value=NULL;
						interface_value=NULL;
						values= NULL;
						memset( parameter,'\0',32);
						return config_type;
					}
					else
						config_type=CONFIG_NONE;
				}

				ptr_parameter = strtok(NULL, ",");
			}
			ptr_parameter=NULL;
			memset(read_buffer,'\0',255);
		}
	}

	free(read_buffer);
	fclose(conf);
	parameter_value=NULL;
	interface_value=NULL;
	values= NULL;
	memset( parameter,'\0',32);
	return config_type;
}

//---------------------------------------------------------------------------

static int set_port_ip(char *ipstr, struct sockaddr_in *sa)
{
	memset(sa, 0, sizeof(struct sockaddr_in) )  ;
	sa->sin_family = AF_INET;
	sa->sin_port = htons(SICCT_PORT);
	sa->sin_addr.s_addr = inet_addr(ipstr) ;
	return 0 ; 
}

void setup_tty(int fd)
{
	struct termios tmp ;

	/* set raw input/ output, 1 second timeout */
	/* set port to 9600, 8E1, no flow, whatever we may need..... */
	tcgetattr(fd, &tmp );
	cfmakeraw(&tmp ) ;
	tmp.c_cflag &= ~CRTSCTS;
	tmp.c_cflag  |= (CLOCAL | CREAD | HUPCL | CS8 | PARENB ) ;
	tmp.c_cflag  &= ~(CRTSCTS | PARODD) ;
	tmp.c_cc[VMIN]  = 0;
	tmp.c_cc[VTIME] = 10;
	cfsetispeed(&tmp, B9600);
	cfsetospeed(&tmp, B9600);
	tcsetattr(fd, TCSANOW, &tmp );
}

char ctinit_com(struct terminal_st *terminal)
{

	int nwritten=0, nread=0 ;
	unsigned char rsp[16] ;

	if( (terminal->port_fd=open(terminal->devname, O_RDWR | O_NOCTTY | O_NDELAY)) < 0) {
			DBG("failed to open %s\n", terminal->devname) ;
			sleep(1);
			return -1;
	} else {
		DBG("device is now open %s\n",terminal->devname) ;
		/* Save current port settings */
		tcgetattr(terminal->port_fd, &(terminal->old_tm) );
		/* set port settings for communication with CardTerminal */
		setup_tty(terminal->port_fd);
		usleep(1000);
	}

	fcntl(terminal->port_fd, F_SETFL, 0);
	usleep(1000);

	/* allow longer timeouts for resync */
	terminal->tm.c_cc[VMIN]  = 0;
	terminal->tm.c_cc[VTIME] = 100;
	tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );

	/* ask resync */
	nwritten=write(terminal->port_fd, ask_resync, 4 );
	if( nwritten < 0 ) {
		DBG("write resync failed\n") ;
		close(terminal->port_fd);
		return CT_ERR_TRANS ;
	}
	usleep(100000);

	nread = read(terminal->port_fd, rsp, 4 );

	if( nread < 0 ) {
		DBG("Read Response Header failed %d\n",nread) ;
		close(terminal->port_fd);
		return CT_ERR_TRANS ;
	}

	if(nread != 4) {
		DBG("Resync Response bad len %d\n", nread) ;
		close(terminal->port_fd);
		return CT_ERR_TRANS ;
	}

	DBG("Read %d bytes\n", nread);
	dump(nread, rsp);

	if( (rsp[0] != 0x21) || (rsp[1] != 0xE0) || calc_csum(4, rsp) ) {
		DBG("Bad Resync Response 0x%.2x 0x%.2x 0x%.2x 0x%.2x\n",
				rsp[0], rsp[1], rsp[2], rsp[3]) ;
		close(terminal->port_fd);
		return CT_ERR_TRANS ;
	}

	/* set default timeout 1 seconds */
	terminal->tm.c_cc[VMIN]  = 0;
	terminal->tm.c_cc[VTIME] = 10;
	tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );

	return CT_OK;
}

static char ctinit_net(struct terminal_st *terminal)
{
	terminal->Socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP) ;
	if(terminal->Socket == INVALID_SOCKET) {
		DBG("Failed to create socket\n") ;
		return CT_ERR_HOST ;  
	}

	if( 0 != connect(terminal->Socket, (struct sockaddr*)&terminal->sa, sizeof(struct sockaddr_in)) ){
		DBG("Failed to connect socket\n") ;
		close(terminal->Socket) ;
		terminal->Socket = INVALID_SOCKET ;
		return CT_ERR_TRANS ;
	}
	return CT_OK ; 
}

char CT_init(unsigned short ctn, unsigned short  pn)
{
	//TODO Sockets;
	char ret ;
	struct terminal_st *term ;
	int new = 0 ;
	int config_type;

	DBG("CT_Open ctn=%d pn=%d\n", (int)ctn, (int)pn ) ;

	if(pn < 1) return CT_ERR_INVALID ;

	// check if this terminal has been opened before
	term = get_terminal(ctn) ;
	if(term) {
		if(   (term->port_fd != INVALID_HANDLE_VALUE)
		   || (term->Socket != INVALID_SOCKET)
		   || (term->pn != pn) ){
			return CT_ERR_INVALID ;
		}
	} else {
		new = 1 ;
		term = new_terminal() ;
	}
	if(!term) return CT_ERR_HOST ;
	term->ctn = ctn ;
	term->pn = pn ;

	config_type = get_config(term);

	if(config_type == CONFIG_NONE) {
		snprintf(term->devname, sizeof(term->devname),"/dev/ttyS%d",pn);
		ret = ctinit_com(term) ;
	} else if (config_type == CONFIG_COM) {
		ret = ctinit_com(term);
	} else if (config_type == CONFIG_IP) {
		if(0 != set_port_ip(term->ip_addr, &term->sa)) ret = CT_ERR_INVALID ;
		else ret = ctinit_net(term) ; 
	} else if (config_type == CONFIG_MAC) {
		int imac[6] ;
		unsigned char mac[6] ;
		char ip[16] ; 
		int res = sscanf(term->mac_addr,"%x:%x:%x:%x:%x:%x",
			         &imac[0], &imac[1], &imac[2], &imac[3], &imac[4], &imac[5]) ;
		if(res != 6) {
			ret = CT_ERR_INVALID ;
		} else {
			int i ;
			for(i=0; i<6; i++) mac[i] = (unsigned char)(imac[i] & 0xFF) ;
			res = find_sicct_term(mac, ip) ;
			if(res) res = (0 == set_port_ip(ip, &term->sa)) ;
			if(res) ret = ctinit_net(term) ;
			else 	ret = CT_ERR_INVALID ;
		}
	} else {
		DBG("Invalid Config %d from Registry\n", config_type) ;
		ret = CT_ERR_INVALID ; 
	}

	term->config_type = config_type;

	if(ret != CT_OK) {
		if(new) free(term) ;
		return ret ;
	}

	if(new) add_terminal(term) ;

	return CT_OK;
}

//---------------------------------------------------------------------------

static char ctdata_com(struct terminal_st *terminal, unsigned char *dad, unsigned char *sad,
               unsigned short lenc, const unsigned char *cmd, unsigned short *lenr,
               unsigned char *rsp )
{

	int nwritten, nread ;
	int max_resp = (int) *lenr ;
	int n ;
	char ret = CT_OK ;
	int sent_total ;
	int total_written ;
	unsigned char rsad, rdad ;
	unsigned char rrseq ;
	unsigned char buf[512]  ; /* realy need 3 byte header, up to 254 byte data + 1Byte ECC */

	sent_total = 0 ;
	while(lenc) {
		int blen ;
		int chain = 0 ;

		buf[0] = (*dad <<4)| *sad; /* NAD Byte */
		buf[1] = terminal->iseq ;
		blen = lenc ;
		if(blen > MAX_TX_LEN) {
			blen = MAX_TX_LEN ;
			buf[1] |= IB_CHAIN ;
			chain = 1 ;
		}
		buf[2] = blen & 0xff ;
		memcpy(&buf[3], &cmd[sent_total], blen) ;
		buf[blen+3] = calc_csum(blen+3, buf) ;

		/* write out block */
		n = blen+4 ;
		total_written = 0 ;
		while(n) {
			nwritten = write(terminal->port_fd, &buf[total_written], n);
			if(  nwritten < 0 ) {
				DBG("data write error\n") ;
				//close(terminal->port_fd) ;
				return CT_ERR_TRANS ;
			}
			n -= nwritten ;
			total_written += nwritten ;
		}

		/* wait for r block */
		if(chain) {
			terminal->tm.c_cc[VMIN]  = 4;
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm));
			nread = read(terminal->port_fd, buf, 4 );
			if( nread < 0) {
				DBG("Read Chain Response failed\n") ;
				//close(terminal->port_fd) ;
				terminal->tm.c_cc[VMIN]  = 0;
				tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm));
				return CT_ERR_TRANS ;
			}
			if(nread < 4) {
				DBG("Read Chain Response short: %d\n", nread) ;
				//close(terminal->port_fd);
				terminal->tm.c_cc[VMIN]  = 0;
				tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm));
				return CT_ERR_TRANS ;
			}
			if( (buf[1] & ~(RB | RB_SEQ)) != 0 ) {
				DBG("Bad R-Block\n") ;
				dump(nread, buf) ;
				terminal->tm.c_cc[VMIN]  = 0;
				tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm));
				//close(terminal->port_fd) ;
				return CT_ERR_TRANS ;
			}
			terminal->tm.c_cc[VMIN]  = 0;
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );
		}
		terminal->iseq ^= IB_SEQ ;
		lenc -= blen ;
		sent_total += blen ;
	}

	*lenr = 0 ;
	rrseq = 0 ;
	while(1) {

		int ntry = 0 ;
		int nread_total = 0 ;
		do {
			terminal->tm.c_cc[VMIN]  = 3-nread_total;
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );
			nread = read(terminal->port_fd, buf+nread_total , 3-nread_total);
			if(nread < 0) {
				DBG("Read Response Header failed\n") ;
				//close(terminal->port_fd);
				return CT_ERR_TRANS ;
			}
			nread_total += nread ;
			if(nread_total < 3) {
				DBG("Response Header bad  Len=%d try=%d\n", nread, ntry);
				ntry ++ ;
				if(ntry > 3) {
					//close(terminal->port_fd);
					return CT_ERR_TRANS ;
				}
			} else break ;
		} while(1) ;
		rsad = buf[0] & 0x0f ;
		rdad = buf[0] >> 4 ;

		/* ToDo : add loop, check nread */
		n = buf[2] + 1 ;
		nread_total = 0 ;
		ntry = 0 ;
		do {
			terminal->tm.c_cc[VMIN]  = n - nread_total;
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );
			nread = read(terminal->port_fd, (&buf[3])+nread_total,n - nread_total );
			if(nread < 0 ) {
				DBG("Read Response Header failed\n") ;
				//close(terminal->port_fd);
				return CT_ERR_TRANS ;
			}
			nread_total += nread ;
			if(nread_total <  n) {
				DBG("Response Data to short %d bytes try = %d\n", nread, ntry) ;
				ntry ++ ;
				if(ntry > 4) {
					//close(terminal->port_fd);
					return CT_ERR_TRANS ;
				}
			} else break ;
		} while(1) ;

		/* ToDo : check sequence */
		if( calc_csum(n+3, buf) ) {
			DBG("Checksum Error\n");
			//close(terminal->port_fd);
			return CT_ERR_TRANS ;
		}

		/* reset timeout to default */
		if(terminal->last_wtx) {
			terminal->tm.c_cc[VMIN]  = 0;
			terminal->tm.c_cc[VTIME] = 10;
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );
			terminal->last_wtx = 0 ;
		}

		/* WTX, terminal request wait */
		if(buf[1] == 0xC3) {
			unsigned char nad ;

			DBG("Got WTX : %d\n", (int)buf[3] );
			nad = buf[0] << 4 ;
			buf[0] = nad | *sad ;
			buf[1] = WTX_RSP ;
			buf[2] = 1 ;
			/* buf[3] = buf[3] ; */
			buf[4] = calc_csum(4, buf) ;
			nwritten = write(terminal->port_fd, buf, 5);
			if( nwritten != 5  ) {
				DBG("WTX Response error\n") ;
				//close(terminal->port_fd);
				return CT_ERR_TRANS ;
			}
			terminal->tm.c_cc[VMIN]  = 0;
			terminal->tm.c_cc[VTIME] = 20 * (int) buf[3];
			tcsetattr(terminal->port_fd, TCSANOW, &(terminal->tm) );
			terminal->last_wtx = 1 ;

		/* continuation block will follow */
		} else if (buf[1] & IB_CHAIN) {
			max_resp -= (int)buf[2] ;
			if(max_resp >= 0) memcpy(&rsp[*lenr], &buf[3], (int) buf[2]) ;
			else ret = CT_ERR_MEMORY ;
			*lenr += (unsigned short)buf[2] ;
			buf[0] = (*dad <<4)| *sad;
			buf[1] = rrseq | RB ;
			buf[2] = 0 ;
			buf[3] = calc_csum(3, buf) ;
			nwritten = write(terminal->port_fd, buf, 4);
			if( nwritten != 4  ) {
				DBG("RRB write error\n") ;
				//close(terminal->port_fd);
				return CT_ERR_TRANS ;
			}
			rrseq ^= RB_SEQ ;
		} else {
			max_resp -= (int) buf[2] ;
			if(max_resp >= 0) memcpy(&rsp[*lenr], &buf[3], (int) buf[2]) ;
			else ret = CT_ERR_MEMORY ;
			*lenr += (unsigned short)buf[2] ;
			break ;
		}
	}

	//DBG("Response len %d\n", (int) *lenr) ;
	//dump((int)*lenr, rsp) ;
	*dad = rdad ;
	*sad = rsad ;

	return ret  ;
}

static int clear_rx(int sk, unsigned int len)
{
	unsigned char buf[256] ;
	int rlen ;

	while(len) {
		if(len > 256) rlen = 256 ;
		rlen = recv(sk, buf, rlen, 0) ;
		if(rlen <= 0) return rlen ;
		len -= rlen ; 
	}
	return 1 ; 
}

static char ctdata_net(struct terminal_st *terminal, unsigned char *dad, unsigned char *sad,
               unsigned short lenc, const unsigned char *cmd, unsigned short *lenr,
               unsigned char *rsp )
{
	unsigned char env[SICCT_ENV_SIZE] ;
	int n ;
	unsigned short tx_dest ;
	unsigned char *txmsg ;

	/* allocate singled buffer to hold cmd + envelop */
	txmsg = malloc(SICCT_ENV_SIZE + (int)lenc) ;
	if(!txmsg) return CT_ERR_MEMORY ;

	/* map dad to dest */
	if( *dad == 0 ) tx_dest = 1 ;
	else if (*dad == 1) tx_dest = 0 ;
	else tx_dest = (unsigned short) *dad ; 

	/* set up and send envelop + apdu */
	txmsg[SICCT_ENV_MESSAGE_TYPE_OFF] = SICCT_MTYPE_CCMD  ;
	set_nbufs(&txmsg[SICCT_ENV_SRC_OR_DEST_OFF], tx_dest) ;
	set_nbufs(&txmsg[SICCT_ENV_SEQ_OFF], terminal->tx_seq) ;
	set_nbufl(&txmsg[SICCT_ENV_LEN_OFF], (unsigned long) lenc) ;
	memcpy(&txmsg[SICCT_ENV_SIZE], cmd, (int) lenc) ;
	n = send(terminal->Socket, txmsg, (int) lenc + SICCT_ENV_SIZE, 0) ;
	free(txmsg);
	if(n != ((int) lenc + SICCT_ENV_SIZE) ) {
		DBG("Socket send failed\n") ;
		close(terminal->Socket) ;
		terminal->Socket = INVALID_SOCKET ;
		return CT_ERR_TRANS ;
	}

	while(1)  {
		int ret ; 
		unsigned long rx_len ;
		unsigned short rx_seq ;
		unsigned short rx_src ; 

		/* receive reply header  */
		n = 0 ; 
		do {
			ret = recv(terminal->Socket, &env[n], SICCT_ENV_SIZE-n, 0) ;
			if(ret <= 0) break ; 
			n  += ret ;
		} while(n < SICCT_ENV_SIZE) ;
		if(ret <= 0){
			DBG("Envelop receive error:%d\n", n) ; 
			break ;
		}

		rx_src = get_nbufs(&env[SICCT_ENV_SRC_OR_DEST_OFF]) ;
		rx_seq = get_nbufs(&env[SICCT_ENV_SEQ_OFF]) ;
		rx_len = get_nbufl(&env[SICCT_ENV_LEN_OFF]) ;

		/* SICCT events not handled yet, silently ignore */ 
		if( env[SICCT_ENV_MESSAGE_TYPE_OFF] == SICCT_MTYPE_EVENT ) {
			DBG("Got sicct event\n") ;
			ret = clear_rx(terminal->Socket, rx_len) ;
			if(ret <= 0) break ; 
		} else if ( env[SICCT_ENV_MESSAGE_TYPE_OFF] == SICCT_MTYPE_RCMD ) {
			if(rx_seq != terminal->tx_seq) {
				DBG("Invalid Sequence in Reply\n") ;
				break ; 
			}
			if(rx_len > (unsigned long)*lenr) {
				DBG("rx packet will not fit\n") ;
				ret = clear_rx(terminal->Socket, rx_len) ;
				if(ret <= 0) break ;
				return CT_ERR_MEMORY ; 
			}

			n = 0 ; 
			do {
				ret = recv(terminal->Socket, &rsp[n], rx_len - n, 0) ;
				if(ret <= 0) break ; 
				n  += ret ;
			} while(n < rx_len) ;
			if(ret <= 0) break ;
			*lenr = (unsigned short) (rx_len & 0xffff) ; 
			*dad = *sad ;
			if(rx_src == 0) *sad = 1 ;
			else if(rx_src == 1) *sad = 0 ;
			else *sad = (unsigned char)(rx_src & 0xff) ;
			return CT_OK ; 
		} else {
			DBG("Invalid Reply Packet type\n") ;
			break ; 
		}
	}
	
	/* if we get here something went wrong, close socket and return error */
	DBG("Closing Socket\n") ;
	close(terminal->Socket) ;
	DBG("Socket closed\n") ;
	terminal->Socket = INVALID_SOCKET ;
	return CT_ERR_TRANS ;
	
}


char CT_data(unsigned short ctn, unsigned char *dad, unsigned char *sad,
				unsigned short lenc, unsigned char *cmd, unsigned short *lenr,
				unsigned char *rsp)
{
	char ret  ;
	struct terminal_st *terminal ;

	DBG("CT_Data ctn=%d, len=%d, sad=0x%x dad=0x%x\n",
		(int)ctn, (int)lenc, (int)*sad, (int)*dad) ;

	if( !print_bc( (int)lenc, cmd) ) {
		dump((int)lenc, cmd) ;
	}

	terminal = get_terminal(ctn) ;
	if(!terminal) {
		return CT_ERR_INVALID ;
	}

	/* check sad, dad */
	if( (*sad != 0x02) && (*sad != 0x05) ) {
		return CT_ERR_INVALID ;
	}
	if(*dad > 0x0E) {
		return CT_ERR_INVALID ;
	}

	/* check len */
	if(lenc < 2) {
		return CT_ERR_INVALID ;
	}

	if(*lenr < 2) {
		return CT_ERR_INVALID ;
	}

	if(terminal->port_fd != INVALID_HANDLE_VALUE) {
		ret = ctdata_com(terminal, dad, sad, lenc, cmd, lenr, rsp) ;
	} else if(terminal->Socket != INVALID_SOCKET) {
		ret = ctdata_net(terminal, dad, sad, lenc, cmd, lenr, rsp) ;
	} else {
		return CT_ERR_INVALID ;
	}

	if(ret == CT_OK) {
		DBG("Response len %d\n", (int) *lenr) ;
		dump((int)*lenr, rsp) ;
	} else {
		DBG("CT_DATA failed %d\n", (int)ret) ;
	}

	return ret;
}

//---------------------------------------------------------------------------

static void CloseCom(struct terminal_st *terminal)
{
	if(terminal->port_fd != INVALID_HANDLE_VALUE) {
		/* restore port settings */
		tcsetattr(terminal->port_fd,TCSANOW, &(terminal->old_tm) );
		close(terminal->port_fd);
	}
	terminal->port_fd = INVALID_HANDLE_VALUE;
}

char CT_close(unsigned short  ctn)
{
	struct terminal_st * terminal;
	unsigned char resp[2] ;
	unsigned char dad, sad ;
	unsigned short lenr;
	DBG("CT_close\n");

	terminal = get_terminal(ctn) ;
	if(!terminal) return CT_ERR_INVALID ;

	dad = 1 ;
	sad = 2 ;
	lenr = 2 ;
	if(terminal->port_fd != INVALID_HANDLE_VALUE) {
		ctdata_com(terminal, &dad, &sad, sizeof(reset_ct), reset_ct, &lenr, resp) ;
		CloseCom(terminal);
	}
	if (terminal->Socket != INVALID_SOCKET) {
		ctdata_net(terminal, &dad, &sad, sizeof(reset_ct), reset_ct, &lenr, resp) ;
		shutdown(terminal->Socket, SHUT_RDWR);
		close(terminal->Socket) ;
		terminal->Socket = INVALID_SOCKET ;
	}
	remove_terminal(ctn) ;

	return CT_OK;
}

